import threading

from ._undname import ffi, lib


class UndnameFailure(Exception):
    pass


def undname(mangled, name_only=False, leading_underscores=True,
            ms_keywords=True, function_returns=True, allocation_language=True,
            thistype=True, access_specifiers=True, member_type=True,
            arguments=True, complex_type=True):
    flags = 0
    if name_only:
        flags += 0x1000
    if not leading_underscores:
        flags += 0x0001
    if not ms_keywords:
        flags += 0x0002
    if not function_returns:
        flags += 0x0004
    if not allocation_language:
        flags += 0x0010
    if not thistype:
        flags += 0x0060
    if not access_specifiers:
        flags += 0x0080
    if not member_type:
        flags += 0x0200
    if not arguments:
        flags += 0x2000
    if not complex_type:
        flags += 0x8000
    _thread_local.undname_exc = None
    try:
        result = lib.undname(mangled.encode('ascii'), flags)
        if _thread_local.undname_exc:
            exc, _thread_local.undname_exc = _thread_local.undname_exc, None
            raise exc
        else:
            return ffi.string(result).decode('ascii')
    finally:
        lib.free(result)


def onerror(exception, exc_value, traceback):
    if not _thread_local.undname_exc:
        _thread_local.undname_exc = exc_value.with_traceback(traceback)


@ffi.def_extern(onerror=onerror)
def error_message(severity, text):
    if severity < 2:
        raise UndnameFailure(ffi.string(text).decode('ascii'))


_thread_local = threading.local()



import unittest
from undname import undname, UndnameFailure


class UndnameTests(unittest.TestCase):

    def test_empty(self):
        res = undname("")

        self.assertEqual("", res)

    def test_unmangled(self):
        res = undname("main")

        self.assertEqual("main", res)

    def test_mangled_complete(self):
        res = undname("?xyz@?$abc@V?$def@H@@PAX@@YAXXZ")

        self.assertEqual(
            "void __cdecl abc<class def<int>,void *>::xyz(void)", res)

    def test_mangled_name_only(self):
        res = undname("?xyz@?$abc@V?$def@H@@PAX@@YAXXZ", name_only=True)

        self.assertEqual(
            "abc<def<int>,void *>::xyz", res)

    def test_raises_on_error(self):
        with self.assertRaises(UndnameFailure):
            undname("?")

    def test_name_only_no_ms_keywords(self):
        res = undname(
            "??B?$PPtr@VLevelGameManager@@@@QEBAPEAVLevelGameManager@@XZ",
            name_only=True, ms_keywords=False)

        self.assertEqual("PPtr<LevelGameManager>::operator LevelGameManager *",
                         res)

    def test_function_in_anonymous_namespace(self):
        res = undname("?ThreadFunc@?A0x6a578590@base@@YAKPEAX@Z", ms_keywords=False)

        self.assertEqual(
            "unsigned long base::`anonymous namespace'::ThreadFunc(void *)",
            res)



from pathlib import Path

import cffi

ffi_builder = cffi.FFI()
ffi_builder.cdef("""
extern void free(void*);
extern char* undname(const char* mangled, unsigned short int flags);
extern "Python+C" void error_message(int severity, const char* message);
""")

ffi_builder.set_source("undname._undname", """
typedef void* (*malloc_func_t)(size_t);
typedef void  (*free_func_t)(void*);

char* __unDName(char *,const char*,int,malloc_func_t,free_func_t,unsigned short int);

char* undname(const char* mangled, unsigned short int flags) {
    return __unDName(NULL, mangled, 0, malloc, free, flags);
}

""", sources=[str(Path(__file__).parent / "src" / "undname.c")])

if __name__ == "__main__":
    ffi_builder.compile(verbose=True)



from setuptools import setup

with open('README.rst') as README:
    README = README.read()

setup(
    name='undname',
    version='0.1.1',
    packages=['undname'],
    url='https://github.com/ikseek/undname',
    license='LGPLv2.1',
    author='Igor Kozyrenko',
    author_email='igor@ikseek.com',
    description='CFFI wrapper for wine undname module',
    long_description=README,
    keywords=['undname', 'msvc'],
    setup_requires=["cffi"],
    install_requires=["cffi"],
    cffi_modules=["undname/undname_build.py:ffi_builder"],
    package_data={'undname': ['src/*.h', 'src/wine/*.h']},
    classifiers=["Development Status :: 4 - Beta",
                 "Topic :: Software Development :: Debuggers",
                 "Programming Language :: Python :: 3.5",
                 "Programming Language :: Python :: 3.6",
                 "License :: OSI Approved"
                 " :: GNU Lesser General Public License v2 or later (LGPLv2+)"]
)
